import logging
import datetime
from typing import List, Optional, Any
from sqlalchemy import func, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship
from marshmallow import fields
from potnanny.models.schemas.safe import SafeSchema
from potnanny.database import Base
from potnanny.plugins.mixins import PluginMixin
from potnanny.models.mixins import CRUDMixin
from potnanny.models.ext import MutableDict, JSONEncodedDict


logger = logging.getLogger(__name__)


class DeviceSchema(SafeSchema):
    name = fields.String()
    notes = fields.String(allow_none=True)
    room_id = fields.Integer(allow_none=True)
    interface = fields.String()
    attributes = fields.Dict(allow_none=True)


class Device(Base, CRUDMixin, PluginMixin):
    __tablename__ = 'devices'

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]
    interface: Mapped[Optional[str]]
    attributes: Mapped[Optional[dict[str, Any]]] = mapped_column(MutableDict.as_mutable(JSONEncodedDict))
    created: Mapped[datetime.datetime] = mapped_column(server_default=func.now())

    # make relationships compatible with asyncio sessions
    __mapper_args__ = {"eager_defaults": True}

    # relationships
    room_id: Mapped[int] = mapped_column(ForeignKey("rooms.id"), nullable=True)
    measurements: Mapped[List['Measurement']] = relationship(
        backref='device',
        cascade='all,delete',
        lazy='subquery')
    schedules: Mapped[List['Schedule']] = relationship(
        backref='device',
        cascade='all,delete',
        lazy='subquery')
    controls: Mapped[List['Control']] = relationship(
        backref='device',
        cascade='all,delete',
        lazy='subquery')


    def __repr__(self):
        return "<Device ({})>".format(self.name)

    def as_dict(self):
        data = {
            'id': self.id,
            'name': self.name,
            'created': self.created.isoformat() + "Z",
            'room_id': self.room_id,
            'interface': self.interface,
            'attributes': self.attributes,
        }

        try:
            data['controls'] = [c.as_dict() for c in self.controls]
        except:
            pass

        try:
            data['schedules'] = [s.as_dict for s in self.schedules]
        except:
            pass

        try:
            mylist = self.plugin.reports
            if ('temperature' in mylist and
                'humidity' in mylist and 'vpd' not in mylist):
                mylist.append('vpd')
            data['reports'] = mylist
        except:
            pass

        try:
            for k, v in self.plugin.__dict__.items():
                if 'plugin' not in data:
                    data['plugin'] = {}

                if k.startswith('_'):
                    continue

                data['plugin'][k] = v
        except Exception as x:
            logger.debug(x)

        return data


    async def poll(self):
        """
        Poll our plugin instance for measurements, if possible
        args:
            none
        returns:
            dict, or None
        """

        if (not hasattr(self.plugin, 'is_pollable') or
            self.plugin.is_pollable is not True):
            return None

        results = {
            'id': self.id,
            'name': self.name,
            'values': {},
        }

        try:
            logger.debug("polling plugin instance: %s" % self.plugin)
            values = await self.plugin.poll()
            results['values'] = values
        except Exception as x:
            logger.warning(str(x))

        return results


    def read_advertisement(self, device, advertisement):
        """
        Use plugin instance to decode the BLE advertisement data
        args:
            the device reporting data
            the advertisement data
        returns:
            dict, or None
        """

        if (not hasattr(self.plugin, 'is_reader') or
            self.plugin.is_reader is not True):
            return None

        results = {
            'id': self.id,
            'name': self.name,
            'values': {},
        }

        try:
            values = self.plugin.read_advertisement(device, advertisement)
            if values:
                results['values'] = values
        except:
            pass

        return results


    async def on(self, outlet=1):
        """
        Switch device ON
        """

        if (not hasattr(self.plugin, 'is_switchable') or
            self.plugin.is_switchable is not True):
            return None

        try:
            result = await self.plugin.on(outlet)
            return result
        except Exception as x:
            logger.warning(str(x))
            return None


    async def off(self, outlet=1):
        """
        Switch device OFF
        """

        if (not hasattr(self.plugin, 'is_switchable') or
            self.plugin.is_switchable is not True):
            return None

        try:
            result = await self.plugin.off(outlet)
            return result
        except Exception as x:
            logger.warning(str(x))
            return None


    async def set_value(self, data):
        """
        *FUTURE*
        Set key/value data on device.
        args:
            - dict
        returns:
        """

        if (not hasattr(self.plugin, 'is_settable') or
            self.plugin.is_settable is not True):
            return None

        if type(data) is not dict:
            logger.warning("Expecting dict but got %s" % type(data))
            return None

        try:
            result = await self.plugin.set_value(data)
            return result
        except Exception as x:
            logger.warning(x)
            return None



import datetime
from markupsafe import escape
from marshmallow import Schema, EXCLUDE, pre_load

class SafeSchema(Schema):
    """
    Subclass of the marshmallow Schema base.
    Excludes unknown fields, and sanitizes all as potential markdown data
    """

    class Meta:
        unknown = EXCLUDE

    @pre_load
    def _sanitize(self, data, **kwargs):
        """
        sanitize all input data, with markupsafe.escape()
        """

        for key, value in data.items():
            if not value:
                continue

            if isinstance(value, dict):
                value = self._sanitize(value)
                continue

            if isinstance(value, datetime.datetime):
                continue

            clean = escape(value)
            data[key] = str(clean)

        return data



import re
import asyncio
import logging
import datetime
from typing import Any, Optional
from sqlalchemy import func, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship
from marshmallow import fields
from potnanny.models.schemas.safe import SafeSchema
from potnanny.database import Base
from potnanny.plugins.mixins import InterfaceMixin
from potnanny.controllers.trigger import (open_action_triggers,
    new_action_trigger)
from potnanny.utils import evaluate
from potnanny.models.mixins import CRUDMixin
from potnanny.models.ext import MutableDict, JSONEncodedDict


logger = logging.getLogger(__name__)


class ActionSchema(SafeSchema):
    name = fields.String()
    device_id = fields.Integer()
    interface = fields.String()
    attributes = fields.Dict(allow_none=True)


class Action(Base, CRUDMixin, InterfaceMixin):
    __tablename__ = 'actions'

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]
    interface: Mapped[Optional[str]]
    attributes: Mapped[Optional[dict[str, Any]]] = mapped_column(MutableDict.as_mutable(JSONEncodedDict))
    created: Mapped[datetime.datetime] = mapped_column(server_default=func.now())

    # make relationships compatible with asyncio sessions
    __mapper_args__ = {"eager_defaults": True}

    # relationships
    device_id: Mapped[int] = mapped_column(ForeignKey("devices.id"))


    def __repr__(self):
        return "<Action ({})>".format(self.name)


    def as_dict(self):
        data = {
            'id': self.id,
            'name': self.name,
            'created': self.created.isoformat() + "Z",
            'device_id': self.device_id,
            'interface': self.interface,
            'attributes': self.attributes,
        }

        return data


    async def _existing_trigger(self):

        try:
            minutes = int(self.attributes['sleep_minutes'])
            now = datetime.datetime.utcnow()
            start = now - datetime.timedelta(minutes=minutes)
            results = await open_action_triggers(self.id, start)
            if results:
                return True
        except:
            pass

        return False


    async def input(self, data):
        """
        Accept measurement data as input, evaluate, and perform actions
        as required:
        args:
            - dict (measurement data from a device)
        returns:
        """

        success = False

        if self._precheck(data) is not True:
            logger.debug(f"data {data} is not acceptable. skipping")
            return

        try:
            equation = "%0.1f %s %s" % (
                float(data['value']),
                self.attributes['condition'],
                self.attributes['threshold'])
            if evaluate(equation) is not True:
                return
        except:
            return

        if await self._existing_trigger():
            logger.debug(f"data {data} has an existing trigger. skipping")
            return

        try:
            klass = self.interface_class(self.interface)
            obj = klass(action_id=self.id, action_name=self.name)
            success = await obj.input(data)
        except Exception as x:
            logger.debug(f"Action interface plugin fail: {x}")
            return

        try:
            if (success and 'sleep_minutes' in self.attributes and
                self.attributes['sleep_minutes'] > 0):
                await new_action_trigger(self.id)
        except Exception as x:
            logger.warning(f"Failed to create action trigger for {self.name}")


    def _precheck(self, data):
        """
        examine incoming data, decide if it is valid for us
        """

        try:
            if data['type'] != self.attributes['type']:
                return False
        except:
            return False

        try:
            if data['device_id'] != self.device_id:
                return False
        except:
            return False

        return True



import logging
import datetime
from typing import Any
from sqlalchemy import func
from sqlalchemy.orm import Mapped, mapped_column, relationship
from marshmallow import fields
from potnanny.models.schemas.safe import SafeSchema
from potnanny.database import Base
from potnanny.models.mixins import CRUDMixin
from potnanny.models.ext import MutableDict, JSONEncodedDict


logger = logging.getLogger(__name__)


class KeychainSchema(SafeSchema):
    name = fields.String()
    attributes = fields.Dict(allow_none=True)
    protected = fields.Boolean(allow_none=True)


class Keychain(Base, CRUDMixin):
    __tablename__ = 'keychains'

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(unique=True)
    attributes: Mapped[dict[str, Any]] = mapped_column(MutableDict.as_mutable(JSONEncodedDict))
    protected: Mapped[bool] = mapped_column(default=False)
    created: Mapped[datetime.datetime] = mapped_column(server_default=func.now())

    __mapper_args__ = {'eager_defaults': True}


    def __repr__(self):
        return "<Keychain({})>".format(self.name)


    def as_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'attributes': self.attributes,
            'protected': self.protected
        }



import logging
import datetime
from sqlalchemy import func, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship
from potnanny.database import Base
from potnanny.models.mixins import CRUDMixin


logger = logging.getLogger(__name__)


class Trigger(Base, CRUDMixin):
    __tablename__ = 'triggers'

    id: Mapped[int] = mapped_column(primary_key=True)
    message: Mapped[str]
    created: Mapped[datetime.datetime] = mapped_column(server_default=func.now())

    # make relationships compatible with asyncio sessions
    __mapper_args__ = {"eager_defaults": True}

    # relationships
    action_id: Mapped[int] = mapped_column(ForeignKey("actions.id"))

    def __repr__(self):
        return "<Trigger(id={}, message={})>".format(self.id, self.message)


    def as_dict(self):
        """
        present object as a dict.
        args:
        returns:
            dict
        """

        data = {
            'id': self.id,
            'message': self.message,
            'created': self.created.isoformat() + "Z",
        }

        return data



import asyncio
import unittest
import datetime
import json
import potnanny.database as db
from unittest import IsolatedAsyncioTestCase
from potnanny.models.room import Room
from potnanny.models.device import Device
from potnanny.models.interface import ObjectInterface


class TestUtilsModel(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)

    async def asyncTearDown(self):
        pass

    async def test_cascade_delete(self):
        # create a room
        room = Room(name='test room')
        await room.insert()

        # add a device to the room
        device = Device(
            name='test device',
            interface='potnanny.bogus.test',
            room_id=room.id)
        await device.insert()

        # delete the room with the obj interface
        ifc = ObjectInterface(Room)
        result = await ifc.delete(room.id)
        print("room deleted")

        # ensure rooms devices were also deleted
        ifc = ObjectInterface(Device)
        devices = await ifc.get_all()
        assert len(devices) == 0


if __name__ == '__main__':
    unittest.main()



import asyncio
import logging
import datetime
from typing import List, Optional
from sqlalchemy import func
from sqlalchemy.orm import Mapped, mapped_column, relationship
from marshmallow import fields
from potnanny.models.schemas.safe import SafeSchema
from potnanny.database import Base
from potnanny.models.mixins import CRUDMixin


logger = logging.getLogger(__name__)


class RoomSchema(SafeSchema):
    name = fields.String()
    notes = fields.String(allow_none=True)


class Room(Base, CRUDMixin):
    __tablename__ = 'rooms'

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]
    notes: Mapped[Optional[str]]
    created: Mapped[datetime.datetime] = mapped_column(server_default=func.now())

    # make relationships compatible with asyncio sessions
    __mapper_args__ = {"eager_defaults": True}

    # relationships
    devices: Mapped[List['Device']] = relationship(
        backref='room',
        cascade='all,delete',
        lazy='subquery')

    def __repr__(self):
        return f"<Room(id={self.id}, name={self.name})>"

    def as_dict(self):
        """
        present object as a dict.
        args:
            - explore: (boolean) recursively explore relationships and include
              their data as well
        returns:
            dict
        """

        data = {
            'id': self.id,
            'name': self.name,
            'notes': self.notes,
            'created': self.created.isoformat() + "Z",
        }
        try:
            data['devices'] = [d.as_dict() for d in self.devices]
        except:
            pass

        return data



import logging
import datetime
from sqlalchemy import func, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship
from marshmallow import fields
from potnanny.models.schemas.safe import SafeSchema
from potnanny.database import Base
from potnanny.models.mixins import CRUDMixin


logger = logging.getLogger(__name__)


class MeasurementSchema(SafeSchema):
    type = fields.String(allow_none=False)
    value = fields.Float(allow_none=False)
    created = fields.Raw(allow_none=True)
    device_id = fields.Integer(allow_none=False)


class Measurement(Base, CRUDMixin):
    __tablename__ = 'measurements'

    id: Mapped[int] = mapped_column(primary_key=True)
    type: Mapped[str]
    value: Mapped[int]
    created: Mapped[datetime.datetime] = mapped_column(server_default=func.now())

    # make relationships compatible with asyncio
    __mapper_args__ = {"eager_defaults": True}

    # relationships
    device_id: Mapped[int] = mapped_column(ForeignKey("devices.id"), nullable=True)

    def as_dict(self):
        return {
            'id': self.id,
            'type': self.type,
            'value': self.value,
            'created': self.created.isoformat() + "Z",
            'device_id': self.device_id,
        }



import asyncio
import logging
import datetime
import potnanny.database as db
from typing import Callable, Awaitable
from sqlalchemy.sql.expression import select
from sqlalchemy.ext.asyncio.session import AsyncAttrs
from potnanny.locks import LOCKS


logger = logging.getLogger(__name__)


class CRUDMixin(object):
    """
    Adds async CRUD functionality to our base sqlalchemy models
    """

    async def insert(self):
        """insert self into db"""

        async def perform():
            async with db.session() as session:
                session.add(self)
                await session.commit()

        await self._execute(perform)


    async def delete(self):
        """delete self (really, an instance of self) from db"""

        async def perform():
            klass = self.__class__
            stmt = select(klass).filter(klass.id == self.id)
            async with db.session() as session:
                results = await session.execute(stmt)
                obj = results.scalars().one_or_none()
                if obj:
                    await session.delete(obj)
                    await session.commit()

        await self._execute(perform)


    async def update(self):
        """update the instance of self into the db"""

        if hasattr(self, 'modified'):
            self.modified = datetime.datetime.utcnow()

        async def perform():
            changes = 0
            klass = self.__class__
            stmt = select(klass).filter(klass.id == self.id)
            async with db.session() as session:
                results = await session.execute(stmt)
                obj = results.scalars().one_or_none()
                if obj:
                    for attr in dir(self):
                        try:
                            current = getattr(self, attr)
                        except:
                            continue

                        if (attr.startswith('_') or
                            isinstance(current, Callable) or
                            isinstance(current, Awaitable) or
                            isinstance(current, AsyncAttrs._AsyncAttrGetitem)):
                            continue

                        old_value = getattr(obj, attr)
                        if current != old_value:
                            setattr(obj, attr, current)
                            changes += 1

                if changes:
                    await session.commit()

        await self._execute(perform)


    async def _execute(self, f):
        """execute the session-based function, within the scope of the lock"""

        if 'db' in LOCKS and LOCKS['db'] is not None:
            async with LOCKS['db']:
                await f()
        else:
            await f()



import asyncio
import unittest
from unittest import IsolatedAsyncioTestCase
import potnanny.database as db
from potnanny.models.room import Room
from potnanny.models.device import Device
from potnanny.models.interface import ObjectInterface



class TestMixins(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)


    async def asyncTearDown(self):
        pass


    async def test_crud_delete(self):
        room = Room(name='test room')
        await room.insert()
        assert room.id > 0

        # test that room was created in db
        ifc = ObjectInterface(Room)
        rooms = await ifc.get_all()
        assert len(rooms) == 1

        # test that the room was deleted from db
        await room.delete()
        rooms = await ifc.get_all()
        assert len(rooms) == 0


    async def test_crud_update(self):
        room = Room(name='room with notes', notes='my test')
        await room.insert()
        assert room.id > 0

        room.notes += " more test"
        await room.update()

        # ensure the update is reflected in the database
        ifc = ObjectInterface(Room)
        obj = await ifc.get_by_name('room with notes')
        assert obj.notes == 'my test more test'


    async def test_crud_dict_update(self):
        device = Device(
            name='test device',
            interface='mytest.foo.bar.Baz',
            attributes={'foo': 'bar'})
        await device.insert()
        assert device.id > 0

        device.attributes['baz'] = 13
        await device.update()

        # ensure the update is reflected in the database
        ifc = ObjectInterface(Device)
        obj = await ifc.get_by_name('test device')
        assert obj.attributes == {'foo': 'bar', 'baz': 13}

if __name__ == '__main__':
    unittest.main()



import re
import asyncio
import logging
import datetime
from typing import Any
from sqlalchemy import func, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship
from marshmallow import fields
from potnanny.models.schemas.safe import SafeSchema
from potnanny.database import Base
from potnanny.controllers.outlet import switch_device_outlet
from potnanny.utils import evaluate
from potnanny.models.mixins import CRUDMixin
from potnanny.models.ext import MutableDict, JSONEncodedDict


logger = logging.getLogger(__name__)


class ControlSchema(SafeSchema):
    name = fields.String()
    device_id = fields.Integer()
    outlet = fields.Integer()
    attributes = fields.Dict(allow_none=True)


class Control(Base, CRUDMixin):
    __tablename__ = 'controls'

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]
    outlet: Mapped[int]
    attributes: Mapped[dict[str, Any]] = mapped_column(MutableDict.as_mutable(JSONEncodedDict))
    created: Mapped[datetime.datetime] = mapped_column(server_default=func.now())

    # make relationships compatible with asyncio sessions
    __mapper_args__ = {"eager_defaults": True}

    # relationships
    device_id: Mapped[int] = mapped_column(ForeignKey("devices.id"))


    def __repr__(self):
        return "<Control ({})>".format(self.name)


    def as_dict(self):
        data = {
            'id': self.id,
            'name': self.name,
            'created': self.created.isoformat() + "Z",
            'device_id': self.device_id,
            'outlet': self.outlet,
            'attributes': self.attributes,
        }

        return data


    async def input(self, data):
        """
        Accept measurement data as input, evaluate, and perform on/off control
        as required:
        args:
            - dict (measurement data
        returns:
        """

        if self._precheck(data) is not True:
            return

        states = {'on': 1, 'off': 0}
        for key in ('on', 'off'):
            if (key not in self.attributes or not self.attributes[key] or
                self.attributes[key]['condition'] in [None, "", "null"] or
                self.attributes[key]['threshold'] in [None, "", "null"]):
                continue

            working = self.attributes[key]
            equation = "%0.1f %s %s" % (
                float(data['value']),
                working['condition'],
                working['threshold'])

            if evaluate(equation) is True:
                state = states[key]

                # switch device to state
                t1 = asyncio.create_task(self.switch_task(
                    self.device_id,
                    self.outlet,
                    state))

                # timed switch? set back to original state after seconds...
                if 'seconds' in working and int(working['seconds']) > 0:
                    t2 = asyncio.create_task(self.switch_task(
                        self.device_id,
                        self.outlet,
                        int(not state),
                        int(working['seconds'])
                    ))


    def _precheck(self, data):
        """
        examine incoming data, decide if it is valid for us
        """

        try:
            if data['device_id'] != self.attributes['input_device_id']:
                return False
        except Exception as x:
            logger.warning(str(x))
            return False

        try:
            if data['type'] != self.attributes['type']:
                return False
        except Exception as x:
            logger.warning(str(x))
            return False

        return True


    async def switch_task(self, device: int, outlet: int, state: int, delay: int = 0):
        try:
            if delay:
                await asyncio.sleep(delay)
                await switch_device_outlet(device, outlet, state)
            else:
                await switch_device_outlet(device, outlet, state)
        except Exception as x:
            logger.warning(x)
